/**
 * @file simple_message_client_1196.c
 * VCS - Tcp/Ip Exercise - client program connects to simple_message_server on a
 * given port, and sends bulletin board messages. After sending, connection is shutdown
 * and response is stored local.
 *
 * @author Sebastian Vogel       <sebastian.vogel@technikum-wien.at>      Matrikelnummer: 0910258067
 * @author Juergen Zornig        <juergen.zornig@technikum-wien.at>       Matrikelnummer: 0910258047
 * @date 2010/12/07
 *
 * @version $Revision: 32 $
 *
 * @todo nothing to do
 *
 * URL: $HeadURL$
 *
 * Last Modified: $Author: sebastian $
 */

/*
 * -------------------------------------------------------------- includes --
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <limits.h>


#define BUFSIZE 512
#define ERROR -1
#define SUCCESS 0

/*
 * --------------------------------------------------------------- globals --
 */
/* Global variable for the programm name */
static const char *szCommand = "(not set)";
static int help_flag = 0;
static struct option long_options[] = {
        {"server",      required_argument,  NULL,               's'},
        {"port",        required_argument,  NULL,               'p'},
        {"user",        required_argument,  NULL,               'u'},
        {"image",       optional_argument,  NULL,               'i'},
        {"message",     required_argument,  NULL,               'm'},
        {"help",        no_argument,        NULL,               'h'},
};

typedef struct message_params
{
    char* server;
    char* port;
    char* user;
    char* imgurl;
    char* message;
} message_params;

/*
 * -------------------------------------------------------------- functions --
 */
int sendRequest(int sockfd, message_params mp);
int getResponse(int sockfd);
void printError(const char *szMessage);
void printUsage(void);
int writeToFile(FILE *fp, char *buffer, int bytesRead);

/**
 * @brief       Main function
 *
 * This function is the main entry point of the program. -
 * it opens a connection to specified server and sends and receives data
 *
 * \param argc the number of arguments
 * \param argv the arguments itselves (including the program name in argv[0])
 *
 * @return    exit status of program
 * @retval    EXIT_FAILURE      Program terminated due to a failure
 * @retval    EXIT_SUCCESS      Program terminated successfully
 *
 */
int main(int argc, char** argv)
{
    int c;
    int option_index = 0;
    message_params mp;

    int sockfd = 0;
    int rv = 0;

    int status = 0;

    struct addrinfo hints, *servinfo, *p;

    /* Store the programm name in szCommand */
    szCommand = argv[0];

    memset(&mp, 0, sizeof(mp));

    mp.imgurl = NULL;
    mp.message = NULL;
    mp.port = NULL;
    mp.server = NULL;
    mp.user = NULL;
    

    if(argc < 2)
    {
        printUsage();
        return (EXIT_SUCCESS);
    }

    while ((c = getopt_long (argc, argv, "s:p:u:i:m:h", long_options, &option_index)) != -1)
    {
        switch (c)
        {
            case 's':
                mp.server = optarg;
                break;
            case 'p':
                mp.port = optarg;
                break;
            case 'u':
                mp.user = optarg;
                break;
            case 'i':
                mp.imgurl = optarg;
                break;
            case 'm':
                mp.message = optarg;
                break;
            case 'h':
            case '?':                
            default:
                help_flag = 1;
                break;
        }
    }


    /*
     * print usage messsage if help_flag is set or last argument is a non-option 
     * argument or if server, port, user or message argument is not set
     */
    if(
            help_flag == 1 ||
            optind < argc ||/*fclose(tx); */ /*kills socket fd */
            mp.message == NULL ||
            mp.port == NULL ||
            mp.server == NULL ||
            mp.user == NULL
        )
    {
        printUsage();
        return (EXIT_SUCCESS);
    }

     /* do connection */
    memset(&hints, 0, sizeof(hints));       
    hints.ai_family = PF_UNSPEC;           
    hints.ai_socktype = SOCK_STREAM;

    if ((rv = getaddrinfo(mp.server, mp.port, &hints, &servinfo)) != 0) {
        printError("getaddrinfo error ");
        printError(gai_strerror(rv));
        exit(EXIT_FAILURE);
    }

    /* loop through the dynamic allocated address info structure to find a matching interface */
    for(p = servinfo; p != NULL; p = p->ai_next )
    {
        if ((sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol)) == -1)
        {
            printError("Couldn't create socket descriptor.");
            continue;
        }

        if (connect(sockfd, p->ai_addr, p->ai_addrlen) == -1)
        {
            close(sockfd);
            printError("Connection failed");
            exit(EXIT_FAILURE);
        }
        /* got connection */
        break;
    }
    /* addrinfo struct isn't needed any more */
    freeaddrinfo(servinfo);

    /* transmission to server */
    if(sendRequest(sockfd, mp) == ERROR)
    {
        close(sockfd);
        exit(EXIT_FAILURE);
    }

    /* send EOF (FIN package) */
    if(shutdown(sockfd, SHUT_WR) == -1)
    {
        close(sockfd);
        printError("Could not shutdown connection");
        exit(EXIT_FAILURE);
    }
    
    /* receive response from server */
    if ((status = getResponse(sockfd)) == ERROR)
    {
        close(sockfd);
        exit(EXIT_FAILURE);
    }
    
    if(close(sockfd)!= 0)
    {
        printError("Could not close connection");
        exit(EXIT_FAILURE);
    }

    exit(EXIT_SUCCESS);
}

/**
 * @brief sendRequest
 *
 * sends the request to the server
 * writes Errors to stderr
 *
 * \param sockfd file descriptor of the socket
 * \param mp message parameters
 *
 * \return int
 * \retval SUCCESS on Success
 * \retval ERROR on Error
 *
 */
int sendRequest(int sockfd, message_params mp)
{
    FILE *tx = NULL;
    
    if((tx = fdopen(sockfd, "w")) == NULL)
    {
        printError("sending: fdopen");
        return ERROR;
    }

    if(fprintf(tx, "user=%s\n", mp.user) < 0)
    {
        printError("sending: fprintf");
        return ERROR;
    }

    if(mp.imgurl != NULL)
    {
        if(fprintf(tx, "img=%s\n", mp.imgurl) < 0)
        {
            printError("sending: fprintf");
            return ERROR;
        }
    }

    if(fprintf(tx, "%s\n", mp.message) < 0)
    {
        printError("sending: fprintf");
        return ERROR;
    }
    
    if(fflush(tx) != 0)
    {
        printError("sending: fflush");
        return ERROR;
    }
    return SUCCESS;
}

/**
 * @brief getResponse
 *
 * receives the response from the server and writes it to theo specified files
 * writes Errors to stderr
 *
 * \param sockfd file descriptor of the socket
 *
 * \return int
 * \retval status returns the status as returned by the server
 * \retval ERROR on Error
 *
 */
int getResponse(int sockfd)
{
    char buf[BUFSIZE];
    int bytesToRead = BUFSIZE;
    int bytesWritten = 0;
    int matches = 0;

    int status = 0;
    char file[NAME_MAX];
    int len = 0;

    int modeParseHeader = 1;
    int modeReadFirstFile = 1;
    
    FILE *fp = NULL;    /* File pointer for reading from socket */
    FILE *fw = NULL;    /* File pointer for writing files */
    
    int status_flag = 0;
    int file_flag = 0;
    int len_flag = 0;
    
    int files_written = 0;
    
    char last_file[NAME_MAX]; 

    
    if((fp = fdopen(sockfd, "r")) == NULL)
    {
        printError("receiving: fdopen");
        return ERROR;
    }

    while(1)
    {   
        if(modeParseHeader == 1)
        {
            while(fgets(buf, bytesToRead, fp) != NULL)
            {
              if((status_flag == 0) && (modeReadFirstFile == 1))
              {
                matches = sscanf(buf,"status=%d", &status);
  
                if ( matches == EOF || matches == 0 )
                {
                    printError("receiving: sscanf");
                    return ERROR; 
                }
                status_flag = 1;
                modeReadFirstFile = 0;
                continue;               
              }
              else if (file_flag == 0)
              {
                matches = sscanf(buf,"file=%s", file);
  
                if ( matches == EOF || matches == 0 )
                {
                    
                    /* 
                     * if this happens when a file has already been written
                     * len as provided by the server was not correct
                     * last file could be corrupt - remove it !
                     */
                    printError("receiving: sscanf");
                    if (files_written > 0)
                    {
                      if(remove(last_file) == -1)
                      {
                        printError("receiving: remove file");
                      }
                      errno = 0;
                      printError("Corrupted transmission");
                    }
                    return ERROR;
                }
                file_flag = 1;
                continue;
              }
              else if (len_flag == 0)
              {
                matches = sscanf(buf,"len=%d", &len);
  
                if ( matches == EOF || matches == 0 )
                {
                    printError("receiving: sscanf");
                    return ERROR;
  
                }
                len_flag = 1;
                if(len < bytesToRead)
                {
                  bytesToRead = len;
                }
                modeParseHeader = 0;
                break;
              }
              else
              {
                /* An error occured */
                errno = 0;
                printError("recieving: corrupt transmission");
                return ERROR;
              }            
            }
            
            if (feof(fp))
                break;

            if (ferror(fp))
            {
                printError("receiving: fgets");
                return ERROR;
            }             
        }
        else
        {
            int bytesRead = 0;
            int eof_flag = 0;
            if((fw = fopen(file, "w")) == NULL)
            {
                printError("receiving: local file store failed");
                return ERROR;
            }

            do
            {
              
              bytesRead = fread(buf, sizeof(char), bytesToRead, fp);
              
              if(bytesRead < bytesToRead)
              {
                eof_flag = 1;
              }
              
              if(bytesRead == 0)
                break;
              
              if(writeToFile(fw, buf, bytesRead) == ERROR)
              {
                  fclose(fw);
                  printError("receiving: writeToFile");
                  return ERROR;
              }
              bytesWritten += bytesRead;
      
              /* get the next bunch of data from socket */
              if(bytesWritten < len)
              {
                  if((len-bytesWritten) > BUFSIZE)
                      bytesToRead = BUFSIZE;
                  else
                      bytesToRead = len - bytesWritten;
              }    
              else
              {
                  /* File written */
                  modeParseHeader = 1;
                  bytesToRead = BUFSIZE;
                  bytesWritten = 0;
                  
                  file_flag = 0;
                  len_flag = 0;
                  
                  files_written++;
                  strncpy(last_file, file, NAME_MAX);
              }
              
            } while((eof_flag != 1) && (modeParseHeader == 0));
            
            if(fclose(fw) != 0)
            {
                printError("receiving: Could not close file");
                return ERROR;
            }

            if (feof(fp))
            {
              if(bytesWritten < len)
              {
                /* 
                 * if this happens when a file has already been written
                 * len as provided by the server was not correct
                 * last file could be corrupt - remove it !
                 */
                if(remove(file) == -1)
                {
                  printError("receiving: remove file");
                }
                errno = 0;
                printError("Corrupted transmission");
                return ERROR;
              }
              
              break;
            }
            
            if (ferror(fp))
            {
                printError("receiving: fread");
                return ERROR;
            }             
        }
    }
    
    return SUCCESS;
}

/**
 *@brief printError
 *
 * recieves errno and prints details
 *
 * \param szMessage additional error message to print
 *
 * \return void
 * \retval void
 */
void printError(const char *szMessage)
{
	if (errno != 0)
	{
		(void) fprintf(stderr, "%s: %s - %s\n", szCommand, szMessage, strerror(errno));
	}
	else
	{
		(void) fprintf(stderr, "%s: %s\n", szCommand, szMessage);
	}
}

/**
 * @brief writeToFile
 *
 * opens a file and writes a string into it
 *
 * \param filename pointer to string which contains the filename
 * \param buffer pointer to the string which should be written into the file
 * \param mode string which contains the mode for the opening of the file e.g.: "a" -> append
 *
 * \return int
 * \retval SUCCES on Success
 * \retval ERROR on Error
 *
 */

int writeToFile(FILE *fp, char *buffer, int bytesRead)
{

    if(fwrite(buffer, sizeof (char), bytesRead, fp) == 0)
    {
        return ERROR;
    }

    if(fflush(fp) != 0)
    {
        return ERROR;
    }

    return SUCCESS;
}

/**
 *@brief printUsage
 *
 * prints small usage info if user tries to use test-pipe in a wrong manner
 *
 * \return void
 * \retval void
 */

void printUsage(void)
{
    /* since this is already a warning, error code of fprintf will not be treated (conforms to BES programming styleguide) */
    (void) fprintf(stderr, "usage: %s options\n", szCommand);
    (void) fprintf(stderr, "options:\n");
    (void) fprintf(stderr, "	-s, --server <server>   full qualified domain name or IP address of the server\n");
    (void) fprintf(stderr, "	-p, --port <port>       well-known port of the server [0..65535]\n");
    (void) fprintf(stderr, "	-u, --user <name>       name of the posting user\n");
    (void) fprintf(stderr, "	-i, --image <URL>       URL pointing to an image of the posting user\n");
    (void) fprintf(stderr, "	-m, --message <message> message to be added to the bulletin board\n");
    (void) fprintf(stderr, "	-h, --help\n");
}


