/**
 * @file simple_message_server_1196.c
 * VCS - Tcp/Ip Exercise - server programm listens on a given port (command
 * argument '-p'), forks off new child processes on incoming tcp-connections and
 * starts the 'simple_message_server_logic'
 *
 * @author Sebastian Vogel       <sebastian.vogel@technikum-wien.at>      Matrikelnummer: 0910258067
 * @author Juergen Zornig        <juergen.zornig@technikum-wien.at>       Matrikelnummer: 0910258047
 * @date 2010/12/07
 *
 * @version $Revision: 32 $
 *
 * @todo nothing to do
 *
 * URL: $HeadURL$
 *
 * Last Modified: $Author: sebastian $
 */

/*
 * -------------------------------------------------------------- includes --
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>

/*
 * -------------------------------------------------------------- defines --
 */

/* error and success */
#define SUCCESS 0
#define ERROR -1
/* Number of pending connection in the server queue */
#define BACKLOG 10 /* how many pending connections queue will hold */
/* command for starting the server logic */
#define START_SERVER_LOGIC "/usr/local/bin/simple_message_server_logic"
#define SERVER_LOGIC_COMMAND "simple_message_server_logic"




/*
 * -------------------------------------------------------------- globals --
 */

/* intialized global for the programm name */
static const char *szCommand = "Not yet set";
static int help_flag = 0;
static struct option long_options[] = {
        {"port",        required_argument,  NULL,               'p'},
        {"help",        no_argument,        &help_flag,         1},
        {0,             no_argument,        NULL,                0}
};

/*
 * -------------------------------------------------------------- functions --
 */

char *getPortNumber(int argc, char * const argv[]);
void printUsage(void);
void startClient(int connect_fd);
void loopAccept(int listen_fd);
void printError(const char *szMessage);

/**
 * @brief sigchldHandler
 *
 * Signal handler function to reap dead childs with waitpid
 *
 * \param signum signalnumber is not used by sigchldHandler();
 *
 * \return void
 * \retval void
 *
 */
void sigchldHandler(int signum)
{
    /* use signum once to supress compiler warnings */
    signum = signum;
    while (waitpid(-1, NULL, WNOHANG) > 0);
}

/**
 * @brief       Main function
 *
 * This function is the main entry point of the program. -
 * it listens on a port provided by argv and accepts incoming tcp-connections
 *
 * \param argc the number of arguments
 * \param argv the arguments itselves (including the program name in argv[0])
 *
 * @return    exit status of program
 * @retval    EXIT_FAILURE      Program terminated due to a failure
 * @retval    EXIT_SUCCESS      Program terminated successfully
 *
 */
int main(int argc, char *argv[])
{

    int listen_sock_fd = 0;
    struct addrinfo hints;
    struct addrinfo *servinfo = NULL;
    struct addrinfo *p = NULL;
    struct sigaction sa;
    int yes = 1;
    int rv = 0;
    char* port = 0;

    szCommand = argv[0];

    if ((port = getPortNumber(argc, argv)) == NULL)
    {
        exit(EXIT_SUCCESS);
    }
    
    memset(&hints, 0, sizeof hints);

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE; /* use my IP */


    if ((rv = getaddrinfo(NULL, port, &hints, &servinfo)) != SUCCESS)
    {
        (void) fprintf(stderr, "%s: getaddrinfo: %s\n", szCommand, gai_strerror(rv));
        exit(EXIT_FAILURE);
    }


    /* loop through all the results and bind to the first we can */
    for (p = servinfo; p != NULL; p = p->ai_next)
    {
        if ((listen_sock_fd = socket(p->ai_family, p->ai_socktype, p->ai_protocol)) == ERROR)
        {
            printError("socket");
            continue;
        }

        if (setsockopt(listen_sock_fd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof (int)) == ERROR)
        {
            printError("setsockopt");
            exit(EXIT_FAILURE);
        }

        if (bind(listen_sock_fd, p->ai_addr, p->ai_addrlen) == ERROR)
        {
            close(listen_sock_fd);
            printError("bind");
            continue;
        }

        break;
    }

    if (p == NULL)
    {
        printError("failed to bind");
        exit(EXIT_FAILURE);
    }

    freeaddrinfo(servinfo); /* all done with this structure */

    if (listen(listen_sock_fd, BACKLOG) == ERROR)
    {
        printError("listen");
        exit(EXIT_FAILURE);
    }

    /* Configure the signal handler to deal with dead child processes */
    sa.sa_handler = sigchldHandler;

    if(sigemptyset(&sa.sa_mask) == ERROR)
    {
      printError("sigemptyset");
      exit(EXIT_FAILURE);
    }
    sa.sa_flags = SA_RESTART;
    /*Start the signal handler*/
    if (sigaction(SIGCHLD, &sa, NULL) == ERROR)
    {
        printError("sigaction");
        exit(EXIT_FAILURE);
    }

    loopAccept(listen_sock_fd);

    if(close(listen_sock_fd) != 0)
    {
        printError("Could not close connection.");
        exit(EXIT_FAILURE);
    }

    exit(EXIT_SUCCESS);
}


/**
 * @brief loopAccept
 *
 * waits for incoming connections on listen_fd
 * forks when connected and starts the client
 *
 * \param listen_fd the file descriptor for the listen socket
 *
 * \return void
 * \retval void
 *
 */
void loopAccept(int listen_fd)
{

    int connect_sock_fd = 0;
    socklen_t sin_size;
    struct sockaddr_storage their_addr; /* connector's address information */

    while (1)
    {

        int fork_result = -1;
        sin_size = sizeof their_addr;
        connect_sock_fd = accept(listen_fd, (struct sockaddr *) & their_addr, &sin_size);

        if (connect_sock_fd == ERROR)
        {
            printError("accept");
            continue;
        }

        fork_result = fork();

        switch (fork_result)
        {

            case ERROR:
                /*forking went wrong */
                printError("fork");
                (void) close(connect_sock_fd);
                (void) close(listen_fd);
                exit(EXIT_FAILURE);
                break;

            case SUCCESS:
                /* All this is the child process */
                /* close listen socket - child does not need this */
                if(close(listen_fd) != 0)
                {
                    printError("Could not close listening socket descriptor");
                    (void) close(connect_sock_fd);
                    exit(EXIT_FAILURE);
                }
                startClient(connect_sock_fd);

                break;

            default:
                 /* this is the parent process */
                 /* the parent does not need the connect socket */
                if(close(connect_sock_fd) != 0)
                {
                    printError("Could not close connection socket");
                    (void) close(listen_fd);
                    exit(EXIT_FAILURE);
                }
                break;
        }
    }
}


/**
 * @brief startClient
 *
 * diverts stdout and stdin to connect_fd
 * overloads process image with the server logic
 *
 * \param connect_fd the file descriptor for the connect socket
 *
 * \return void
 * \retval void
 *
 */
void startClient(int connect_fd)
{
    /* divert stdin and stdout to connect socket */
    if ((dup2(connect_fd, STDIN_FILENO) == ERROR) ||
        (dup2(connect_fd, STDOUT_FILENO) == ERROR)
         )
    {
        printError("child: dup");
        (void) close(connect_fd);
        exit(EXIT_FAILURE);
    }

    /* After dup, we dont need the filedescriptor any more -> close it */
    if(close(connect_fd) != 0)
    {
        printError("child: Could not close connect socket");
        exit(EXIT_FAILURE);
    }


    /* Overload child process image */
    (void) execl(START_SERVER_LOGIC, SERVER_LOGIC_COMMAND, (char *) NULL);

    /* this part should not be reached if everything goes well */
    printError("child: execl");
    (void) close(connect_fd);
    exit(EXIT_FAILURE);

}

/**
 * @brief getPortIndexNumber
 *
 * parses the arguments array with getopt and returns index of the position of
 * the portnumber-string in argv
 *
 * \param argc the number of arguments
 * \param argv the arguments itselves (including the program name in argv[0])
 *
 * \return int
 * \retval port_index index of the portnumber in the argv array
 * \retval ERROR on error or if usage message should be displayed
 *
 */
char *getPortNumber(int argc, char * const argv[])
{
        int c = 0;
        char *port = NULL;
        int option_index = 0;


        /* return NULL if the programm name is the only argument */
        if (argc < 2)
        {
            printUsage();
            return NULL;
        }

        /* if opterr == 0 getopt() does not print errors to stderr */
        opterr = 0;

        /* Parse Arguments - on error */
        while ((c = getopt_long (argc, argv, "p:h", long_options, &option_index)) != -1)
        {
            switch (c)
            {
                case 'p':
                    port = optarg;
                    break;
                case 0:
                case 'h':
                case '?':
                default:
                    help_flag = 1;
                    break;
            }
         }


        /* return NULL if help_flag is set or last argument is a non-option argument */
        if (help_flag == 1 || optind < argc || port == NULL)
        {
            printUsage();
            return NULL;
        }

        return port;

}
/**
 * @brief printUsage
 *
 * prints short usage message stdout
 *
 * \return void
 * \retval void
 *
 */
void printUsage(void)
{
     /*
      * since this is already a warning, error code of fprintf will not be
      * treated (conforms to BES programming styleguide)
      */
    (void) fprintf(stderr, "usage: %s option\n", szCommand);
    (void) fprintf(stderr, "options:\n");
    (void) fprintf(stderr, "	-p, --port <port>\n");
    (void) fprintf(stderr, "	-h, --help\n");
}


/**
 *@brief printError
 *
 * recieves errno and prints details
 *
 * \param szMessage additional error message to print
 *
 * \return void
 * \retval void
 *
 */
void printError(const char *szMessage)
{
    (void) fprintf(stderr, "%s: ", szCommand);
    perror(szMessage);

}

/*
 * =================================================================== eof ==
 */


