#include <simple_message_client_commandline_handling.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#define MAX_SIZE 255

int recResponse(int sd, int* verbose, const char* cmnd);
int sendQuery(int sd, const char* user, const char* message, const char* img_url, int* verbose, const char* cmnd);
void log_message(const char* cmnd, int* tmp_verbose, const char* message, int useErrno);
void usage(FILE *stream, const char *cmnd, int exitcode);


int main( int argc, const char* argv[] )
{
	/* variables */
	const char* server = NULL;
	const char* port = NULL;
	const char* user = NULL;
	const char* message = NULL;
	const char* img_url = NULL;
	int verbose = FALSE;
	int result;
	int sd;


	struct addrinfo *server_addr_info;

	/* parse commandline */
	smc_parsecommandline(argc, argv, &usage, &server, &port, &user, &message, &img_url, &verbose);

	/* check for ipv4 or ipv6 */
	result = getaddrinfo(server,port,NULL,&server_addr_info);
	if (result!=0) {	
		log_message(argv[0],&verbose,gai_strerror(result),FALSE);
		
		/* exit TODO */
		return EXIT_FAILURE;
	}

	/* create socket */
	sd=socket(server_addr_info->ai_family,server_addr_info->ai_socktype,server_addr_info->ai_protocol);
	if (sd == -1) {
		log_message(argv[0],&verbose,"Failed to create socket",TRUE);
	
		/* exit TODO */
		return EXIT_FAILURE;
	}

	if (connect(sd,server_addr_info->ai_addr,server_addr_info->ai_addrlen) != 0 ) {
		log_message(argv[0],&verbose,"Failed to connect",TRUE);

		/* exit TODO */
		return EXIT_FAILURE;
	}

	/* clean up getaddrinfo */
        freeaddrinfo(server_addr_info); 
	

	result = sendQuery(sd,user,message,img_url,&verbose,argv[0]);
	if (result != TRUE) {
		(void)close(sd);
		log_message(argv[0],&verbose,"Error while sending query",FALSE);
		return EXIT_FAILURE;
	}

	result = recResponse(sd,&verbose,argv[0]);
	if (result != TRUE) {
		(void)close(sd);
		log_message(argv[0],&verbose,"Error while recieving response",FALSE);
		return EXIT_FAILURE;
	}

	/* close socket and ignore errors */
	(void)close(sd);
	
	/* exit successfully */	
	return EXIT_SUCCESS;
}

int recResponse(int sd, int* verbose, const char* cmnd) {
	
	FILE* sfp;	/* socket file pointer */
	int result;
	char buf[MAX_SIZE];
	int max_buf_size; 
	
	enum states { header, file } state;
	struct metadatas {
		int exitcode;
	} metadata;

	enum file_states { name, len, data  } file_state;
	struct file_metadata {
		FILE* ffp;
		char* name;
		int len;
		int recieved;
	} file_metadata;
	
	/* initialisation */
	char tmp_name[MAX_SIZE];
	file_metadata.name=tmp_name;
	state=header;

	sfp = fdopen(sd,"r");
	if (sfp==NULL) {
		log_message(cmnd,verbose,"Failed to open socket for reading",TRUE);
                return FALSE;
	}

	while(fgets(buf,MAX_SIZE,sfp) != NULL) {
	
		/* expecting header */
		if (state==header) {
			result=sscanf(buf,"status=%d",&metadata.exitcode);
			
			if (result!=1) {
				log_message(cmnd,verbose,"Unexpected input (expected: status=<exitcode>)",TRUE);
                		return FALSE;
			}
			
			state=file;
			file_state=name;			
		}
		/* expecting a file */
		else {
			
			if (file_state==name) {
				result=sscanf(buf,"file=%s",file_metadata.name);
                        	
                if (result!=1) {         
					log_message(cmnd,verbose,"Unexpected input (expected: file=<filename>)",TRUE);
                  	return FALSE;
              	}

				file_metadata.ffp = fopen(file_metadata.name,"w");
				if (file_metadata.ffp==NULL) {
					log_message(cmnd,verbose,"Unable to open file",TRUE);
                    return FALSE;
				}
				file_state=len;

			}
			else if (file_state==len) {
				result=sscanf(buf,"len=%d",&file_metadata.len);
				
				if (result!=1) {
					(void)fclose(file_metadata.ffp); /* close file and ignore errors */
					log_message(cmnd,verbose,"Unexpected input (expected: len=<filelen>)",TRUE);
					return FALSE;
				}
				
				file_metadata.recieved=0;
				file_state=data;
			}
			if (file_state==data) {
		
				do {
					/* check for max size */
					if (MAX_SIZE<(file_metadata.len-file_metadata.recieved)) {
						max_buf_size=MAX_SIZE;
					}
					else {
						max_buf_size=file_metadata.len-file_metadata.recieved;
					}
				
					result=fread(buf,sizeof(char),max_buf_size,sfp);
					
					file_metadata.recieved+=result;
					
					/* write to file */
                    if ( (int)fwrite(buf,sizeof(char),result,file_metadata.ffp) != result) {
                       	(void)fclose(file_metadata.ffp); /* close file and ignore errors */
                        log_message(cmnd,verbose,"Error while writing to file",FALSE);
                        return FALSE;
                    }

					
					/* successfully recieved */
					if (file_metadata.recieved==file_metadata.len) {
						
						if (fclose(file_metadata.ffp)!=0) {
                                	                log_message(cmnd,verbose,"Error while closing file",TRUE);
                        	                        return FALSE;
                	                        }
						
						file_state=name; /* finished */

						break;
					}
				
					if (feof(sfp)!=0 || ferror(sfp)!=0 ) {
						log_message(cmnd,verbose,"Error while reading from stream (connection closed by peer/error)",FALSE);
                                        	return FALSE;
                                        }
					
				} while(1);
			}

		}

	}
	
	if (state==file && file_state==name) {
		return TRUE;
	}
	else {	
		log_message(cmnd,verbose,"Error while reading from stream (connection closed by peer/error)",FALSE);
	
		return FALSE;
	}

}

int sendQuery(int sd, const char* user, const char* message, const char* img_url, int* verbose, const char* cmnd ) {

	FILE* sfp;

	sfp = fdopen(sd,"w");	
	if (sfp==NULL) {
		log_message(cmnd,verbose,"Failed to open socket for writing",TRUE);
		return FALSE;
	}
	
	if (fprintf(sfp,"user=%s\n",user)<0) {
		log_message(cmnd,verbose,"Failed to write user to stream",FALSE);
		return FALSE;
	}

	if (img_url!=NULL) {
		if (fprintf(sfp,"img=%s\n",img_url)<0) {
                	log_message(cmnd,verbose,"Failed to write image url to stream",FALSE);
                	return FALSE;
		}
        }

        if (fprintf(sfp,"%s\n",message)<0) {
                log_message(cmnd,verbose,"Failed to write message to stream",FALSE);
                return FALSE;
        }


	if (fflush(sfp)!=0) {
		log_message(cmnd,verbose,"Failed to flush stream after writing",TRUE);
                return FALSE;
	}

	if (shutdown(sd,SHUT_WR)!=0) {
		log_message(cmnd,verbose,"Failed to close writing direction of stream",TRUE);
                return FALSE;
	}

	return TRUE;
}

	
void log_message(const char* cmnd, int* tmp_verbose, const char* message, int useErrno) {
	FILE *stream=stderr;

	int verbose = *tmp_verbose;

	/* check if verbose */
	if (verbose == TRUE) {
		stream=stdout;
	}
	
	if (useErrno == FALSE) {
		fprintf(stream,"%s: %s\n",cmnd,message);
	}
	else {
		fprintf(stream,"%s: %s (%s)\n",cmnd,message,strerror(errno));
	}

}

/**
 *@brief usage
 *
 * prints small usage info if user tries to use test-pipe in a wrong manner
 *
 * \return void
 * \retval void
 */
void usage(FILE *stream, const char *cmnd, int exitcode) {
	fprintf(stream,"usage: %s options\n",cmnd);
	fprintf(stream,"options:\n");
	fprintf(stream,"-s, --server <server>   full qualified domain name or IP address of the server\n");
	fprintf(stream,"-p, --port <port>       well-known port of the server [0..65535]\n");
	fprintf(stream,"-u, --user <name>       name of the posting user\n");
	fprintf(stream,"-i, --image <URL>       URL pointing to an image of the posting user\n");
	fprintf(stream,"-m, --message <message> message to be added to the bulletin board\n");
	fprintf(stream,"-v, --verbose           verbose output\n");
	fprintf(stream,"-h, --help\n");
	
	exit(exitcode);
}



