#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>
#include <string.h>
#include <libgen.h>

#define TRUE (1==1)
#define FALSE (!TRUE)

#define BACKLOG 10
#define SERVER_LOGIC_CMND "/usr/local/bin/simple_message_server_logic"

typedef void (* sms_usagefunc_t) (FILE *, const char *, int);

void server_logic(int cd, const char* cmnd, int* verbose);
void sigchld_handler(int s);
void usage(FILE *stream, const char *cmnd, int exitcode);
void sms_parsecommandline(int argc, const char* const argv[], sms_usagefunc_t usagefunc ,const char **port, int* verbose);
void log_message(const char* cmnd, int* tmp_verbose, const char* message, int useErrno);


int main( int argc, const char* argv[] ) {

	int result;
	int sd,cd;
	socklen_t sin_size;
	struct addrinfo hints;
	struct addrinfo *server_addr_info, *p;
	struct sigaction sa;	
	struct sockaddr_storage client_addr;

	const char* port = NULL;
	int yes = 1;
	int verbose = FALSE;

	sms_parsecommandline(argc, argv, &usage, &port, &verbose); 

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;	/* prefer inet4 */
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	
	result = getaddrinfo(NULL, port, &hints, &server_addr_info);
	if (result!=0) {	
		log_message(argv[0],&verbose,gai_strerror(result),FALSE);
		
		/* exit TODO */
		return EXIT_FAILURE;
	}

	/* loop through all the results and bind to the first we can */ 
	for(p = server_addr_info; p != NULL; p = p->ai_next) {
	
		if ((sd = socket(p->ai_family, p->ai_socktype, p->ai_protocol)) == -1) {
			log_message(argv[0],&verbose,"Failed to create socket",TRUE);
			continue;
		}

		if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
			log_message(argv[0],&verbose,"Failed to set socket options",TRUE);
			return EXIT_FAILURE;
		}

		if (bind(sd, p->ai_addr, p->ai_addrlen) == -1) { 
			close(sd);
			log_message(argv[0],&verbose,"Failed to bind to socket",TRUE);
			continue;
		} 

		break;
	}

	freeaddrinfo(server_addr_info);

	if (p == NULL) {
		log_message(argv[0],&verbose,"No socket found to bind to",FALSE); 
		return EXIT_FAILURE;
	}

	if (listen(sd, BACKLOG) == -1) { 
		log_message(argv[0],&verbose,"Failed to listen",TRUE);
		return EXIT_FAILURE;
	}
	
	sa.sa_handler = sigchld_handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags=SA_RESTART;
	
	if (sigaction(SIGCHLD, &sa, NULL) == -1) {
		log_message(argv[0],&verbose,"Failed to set sigchld handler",TRUE);
		return EXIT_FAILURE;
	}

	while(1) {
		sin_size = sizeof(client_addr);
		cd = accept(sd, (struct sockaddr *)&client_addr, &sin_size);
		if (cd == -1) {
			log_message(argv[0],&verbose,"Error while accepting new client",TRUE);
			continue;
		}

		result=fork();

		switch(result) {
			case -1:
				(void)close(sd);
				(void)close(cd);
				break;
			case 0:
				/* child */
				if (close(sd) != 0) {
					log_message(argv[0],&verbose,"Error while closing client socket after fork",TRUE);
					(void)close(cd);
					return EXIT_FAILURE;
				}

				server_logic(cd, argv[0], &verbose);

				break;
			default: 
				/* parent */
				if (close(cd) != 0) {
					log_message(argv[0],&verbose,"Error while closing server socket after fork",TRUE);
					(void)close(sd);
					return EXIT_FAILURE;
				}
				break;
		}
	}	
	
	(void)close(sd);
	return EXIT_SUCCESS;
}

void server_logic(int cd, const char* cmnd, int* verbose) {
	
	int result;
	char* slc_copy;
	char* arg0;

	/* redirect stdout and stdin */
	if ( dup2(cd, STDIN_FILENO) != -1 && dup2(cd, STDOUT_FILENO) != 1 ) {
                log_message(cmnd,verbose,"Error while redirecting stdin and stdout",TRUE);
		(void)close(cd);
		exit(EXIT_FAILURE);;
	} 

	if (close(cd) != 0) {
		log_message(cmnd,verbose,"Error while closing client socket after redirecting",TRUE);
                exit(EXIT_FAILURE);
	}

	slc_copy = strdup(SERVER_LOGIC_CMND);
	if (slc_copy == NULL) {
		log_message(cmnd,verbose,"Error while dublicating SERVER_LOGIC_CMND",FALSE);
                exit(EXIT_FAILURE);
	}
	
	arg0 = basename(slc_copy);	

	result = execl(SERVER_LOGIC_CMND, arg0, NULL);
	
	/* shouldn't be reached */
	log_message(cmnd,verbose,"Error while executing SERVER_LOGIC",TRUE);
	(void)close(cd);
	exit(EXIT_FAILURE);
}

void sigchld_handler(int s) {
	s=s; /* hide unused parameter warning */
	while (waitpid(-1, NULL, WNOHANG) > 0);
}


void log_message(const char* cmnd, int* tmp_verbose, const char* message, int useErrno) {
	FILE *stream=stderr;

	int verbose = *tmp_verbose;

	/* check if verbose */
	if (verbose == TRUE) {
		stream=stdout;
	}
	
	if (useErrno == FALSE) {
		fprintf(stream,"%s: %s\n",cmnd,message);
	}
	else {
		fprintf(stream,"%s: %s (%s)\n",cmnd,message,strerror(errno));
	}

}


void usage(FILE *stream, const char *cmnd, int exitcode) {
	fprintf(stream,"usage: %s option\n",cmnd);
	fprintf(stream,"options:\n");
	fprintf(stream,"	-p, --port <port>\n");
	fprintf(stream,"	-v, --verbose\n");
	fprintf(stream,"	-h, --help\n");

	exit(exitcode);

}

void sms_parsecommandline(int argc, const char* const argv[], sms_usagefunc_t usagefunc ,const char **port, int* verbose) {

	int c;
	
	struct option long_options[] = {
		{"port",1,NULL,'p'},
		{"verbose",0,NULL,'v'},
		{"help",0,NULL,'h'},
		{0,0,0,0}
	};

	while( (c = getopt_long(argc, (char ** const) argv, "p:hv", long_options, NULL)) != -1) {

		switch (c) {
			case 'p':
				*port = optarg;
				break;
			case 'v':
				*verbose = TRUE;
				break;
			case 'h': 
				usagefunc(stdout, argv[0], EXIT_SUCCESS);
				break;
			case '?':
			default:
				usagefunc(stderr, argv[0], EXIT_FAILURE);
				break;
		}
	}
	
	if (optind != argc || *port==NULL) {
		usagefunc(stderr, argv[0], EXIT_FAILURE);
	}

}


