/**
 * @mainpage Example Program for TCP/IP Programming
 * @author florian.bogner@technikum-wien.at
 * @author michal.piatek@technikum-wien.at
 */

/**
 * @brief the client from the TCP/IP programming example
 * @file simple_message_client_1384.c
 */

/*
 * -------------------------------------------------------------- includes --
 */
 
#include <simple_message_client_commandline_handling.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

/*
 * --------------------------------------------------------------- defines --
 */

/* maximal size of arrays (and chars) */
#define MAX_SIZE 255

/*
 * ------------------------------------------------------------ prototypes --
 */

int sendQuery(int sd, const char* user, const char* message, const char* img_url, int* verbose, const char* cmnd); 
int recResponse(int sd, int* verbose, const char* cmnd);
void log_verbose(const char* cmnd, int* tmp_verbose, const char* message);
void log_message(const char* cmnd, int* tmp_verbose,int useErrno, const char* message);
void usage(FILE *stream, const char *cmnd, int exitcode);

/*
 * ------------------------------------------------------------- functions --
 */
 
/**
 * @brief the main function, starting point
 *
 * The main function parses the programm arguments, tries to sets up the client socket
 * and executes the send and receive specific functions.
 * 
 * @param  argc argument counter
 * @param  argv[] The arguments passed to the main function.
 * @return EXIT_SUCCESS (0)
 * @return EXIT_FAILURE (1)
 */
int main( int argc, const char* argv[] ) {
	/* local variables */
	const char* server = NULL;
	const char* port = NULL;
	const char* user = NULL;
	const char* message = NULL;
	const char* img_url = NULL;
	int verbose = FALSE;
	int result;
	int sd;

	struct addrinfo *server_addr_info;

	/* parse commandline - arguments set to NULL to work around undocumented behaviour in library */
	smc_parsecommandline(argc, argv, &usage, &server, &port, &user, &message, &img_url, &verbose);

	/* use getaddrinfo to check if IPv4 or IPv6 */
	log_verbose(argv[0],&verbose, "Use getaddrinfo to check if IPv4 or IPv6");
	result = getaddrinfo(server,port,NULL,&server_addr_info);
	if (result!=0) {	
		log_message(argv[0],&verbose,FALSE,gai_strerror(result));
		
		return EXIT_FAILURE;
	}

	/* create client socket */
	log_verbose(argv[0],&verbose, "Create client socket");
	sd=socket(server_addr_info->ai_family,server_addr_info->ai_socktype,server_addr_info->ai_protocol);
	if (sd == -1) {
		log_message(argv[0],&verbose,TRUE,"Failed to create socket");
	
		return EXIT_FAILURE;
	}

	/* try to connect to server */
	log_verbose(argv[0],&verbose, "Try to connect to server");
	if (connect(sd,server_addr_info->ai_addr,server_addr_info->ai_addrlen) != 0 ) {
		log_message(argv[0],&verbose,TRUE,"Failed to connect");

		return EXIT_FAILURE;
	}

	/* clean up getaddrinfo */
	log_verbose(argv[0],&verbose, "Clean up server_addr_info");
	freeaddrinfo(server_addr_info); 
	
	/* send query to server */
	log_verbose(argv[0],&verbose, "Send query to server");
	result = sendQuery(sd,user,message,img_url,&verbose,argv[0]);
	if (result != TRUE) {
		(void)close(sd);
		log_message(argv[0],&verbose,FALSE,"Error while sending query");
		return EXIT_FAILURE;
	}

	/* wait for response and parse it */
	log_verbose(argv[0],&verbose, "Wait for response and parse it");
	result = recResponse(sd,&verbose,argv[0]);
	if (result != TRUE) {
		(void)close(sd);
		log_message(argv[0],&verbose,FALSE,"Error while receiving response");
		return EXIT_FAILURE;
	}

	log_verbose(argv[0],&verbose, "Close socket end exit successfully");
	/* close socket and ignore errors */
	(void)close(sd);
	
	/* exit successfully */	
	return EXIT_SUCCESS;
}

/**
 * @brief sends the message to the server
 *
 * This functions converts the socket descriptor to a FILE* and sends the query to the server using fprintf. 
 * Afterwards the connection is closed for writing.
 * ATTENTION: all variables expect img_url should have a value different to NULL
 *
 * @param sd the socket descriptor of type int referring to the socket file
 * @param user the user data 
 * @param message the message sent 
 * @param img_url the image sent 
 * @param verboseflag
 * @param cmnd command name used for logging
 * @return TRUE
 * @return FALSE
 */
int sendQuery(int sd, const char* user, const char* message, const char* img_url, int* verbose, const char* cmnd ) {
	FILE* sfp;
	
	/* open socket for writing */
	log_verbose(cmnd,verbose, "Open socket for writing");
	sfp = fdopen(sd,"w");	
	if (sfp==NULL) {
		log_message(cmnd,verbose,TRUE,"Failed to open socket for writing");
		return FALSE;
	}
	
	/* write user id */
	log_verbose(cmnd,verbose, "Send user");
	if (fprintf(sfp,"user=%s\n",user)<0) {
		log_message(cmnd,verbose,FALSE,"Failed to write user to stream");
		return FALSE;
	}

	/* write image url (if set) */
	if (img_url!=NULL) {
		log_verbose(cmnd,verbose, "Send image url");
		if (fprintf(sfp,"img=%s\n",img_url)<0) {
                	log_message(cmnd,verbose,FALSE,"Failed to write image url to stream");
                	return FALSE;
		}
    }

	/* write message */
	log_verbose(cmnd,verbose, "Send message");
   	if (fprintf(sfp,"%s\n",message)<0) {
		log_message(cmnd,verbose,FALSE,"Failed to write message to stream");
		return FALSE;
   	}

	/* flush the file* */
	log_verbose(cmnd,verbose, "Flush and close socket for writing");
	if (fflush(sfp)!=0) {
		log_message(cmnd,verbose,TRUE,"Failed to flush stream after writing");
        return FALSE;
	}

	/* shutdown writing direction of socket */
	if (shutdown(sd,SHUT_WR)!=0) {
		log_message(cmnd,verbose,TRUE,"Failed to close writing direction of stream");
        return FALSE;
	}

	return TRUE;
}

/**
 * @brief deals with the answer
 *
 * This function converts the socket descritor to a FILE* and waits for incoming data.
 * Afterwards it tries to read data form the socket according the protocol (see protocol description)
 * If everything went fine, all the sent files will be created in the current working directory.
 *
 * @param sd socket descriptor,
 * @param verboseflag
 * @param command name used for logging
 * @return TRUE if operation succeeds FALSE if operation fails
 */
int recResponse(int sd, int* verbose, const char* cmnd) {
	FILE* sfp;
	int result;
	char buf[MAX_SIZE];
	char tmp_name[MAX_SIZE];
	int max_buf_size; 
	
	/* 
	 * baby state machine for checking main operation
	 * it is used to distinguish between header and file specific data
	 */
	enum states { header, file } state;
	struct metadatas {
		int exitcode;
	} metadata;

	/* 
	 * baby state machine for checking file operations
	 * is is used to distinguish between filename, data lenth and data
	 */
	enum file_states { name, len, data  } file_state;
	struct file_metadata {
		FILE* ffp;
		char* name;
		int len;
		int received;
	} file_metadata;
	
	/* initialisation */
	file_metadata.name=tmp_name;
	state=header;

	/* open socket for reading */
	sfp = fdopen(sd,"r");
	if (sfp==NULL) {
		log_message(cmnd,verbose,TRUE,"Failed to open socket for reading");
                return FALSE;
	}

	/* try to read next line from socket (blocking) */
	log_verbose(cmnd,verbose, "Waiting for data");
	while(fgets(buf,MAX_SIZE,sfp) != NULL) {
	
		/* if we are expecting the header try to read it */
		if (state==header) {
			log_verbose(cmnd,verbose, "Expecting header");
			/* read helper exit code from socket */
			result=sscanf(buf,"status=%d",&metadata.exitcode);
			if (result!=1) {
				log_message(cmnd,verbose,TRUE,"Unexpected input (expected: status=<exitcode>)");
                return FALSE;
			}
			
			if (metadata.exitcode != 0) {
				log_message(cmnd,verbose,FALSE,"Unexpected helper exit code received");
                return FALSE;
			}
			
			/* move to next state */
			log_verbose(cmnd,verbose, "Header received");
			state=file;
			file_state=name;			
		}
		/* if we are expecting file data  */
		else {
			/* if we are expecting the file name, read it and try to open the file */
			
			if (file_state==name) {
				log_verbose(cmnd,verbose, "Expecting filename");
				/* read file name from socket */
				result=sscanf(buf,"file=%s",file_metadata.name);	
                if (result!=1) {         
					log_message(cmnd,verbose,TRUE,"Unexpected input (expected: file=<filename>)");
                  	return FALSE;
              	}

				/* ... and open it */
				file_metadata.ffp = fopen(file_metadata.name,"w");
				if (file_metadata.ffp==NULL) {
					log_message(cmnd,verbose,TRUE,"Unable to open file");
                    return FALSE;
				}
				
				/* move to next state */
				log_verbose(cmnd,verbose, "Filename received");
				file_state=len;
			}
			/* we are expecting the data length */
			else if (file_state==len) {
				log_verbose(cmnd,verbose, "Expecting date length");
				/* read expected data lenght from socket */
				result=sscanf(buf,"len=%d",&file_metadata.len);
				if (result!=1) {
					(void)fclose(file_metadata.ffp); /* close file and ignore errors */
					log_message(cmnd,verbose,TRUE,"Unexpected input (expected: len=<filelen>)");
					return FALSE;
				}
				
				/* move to next state and reset received data */
				log_verbose(cmnd,verbose, "Data length received");
				file_metadata.received=0;
				file_state=data;
			}
			
			/* read the data from the stream */
			if (file_state==data) {
				/* read until all data is received */
				log_verbose(cmnd,verbose, "Expecting data");
				do {
					/* match buffer size against expected data */
					if (MAX_SIZE<(file_metadata.len-file_metadata.received)) {
						max_buf_size=MAX_SIZE;
					}
					else {
						max_buf_size=file_metadata.len-file_metadata.received;
					}
					
					/* read data from stream */
					result=fread(buf,sizeof(char),max_buf_size,sfp);
					
					/* and update received length */
					file_metadata.received+=result;
					
					/* write to file */
                    if ( (int)fwrite(buf,sizeof(char),result,file_metadata.ffp) != result) {
                       	(void)fclose(file_metadata.ffp); /* close file and ignore errors */
                        log_message(cmnd,verbose,FALSE,"Error while writing to file");
                        return FALSE;
                    }

					/* if all data is successfully received, close file */
					if (file_metadata.received==file_metadata.len) {
						
						/* close file */
						if (fclose(file_metadata.ffp)!=0) {
							log_message(cmnd,verbose,TRUE,"Error while closing file");
							return FALSE;
						}
						
						/* ... and reset state */
						file_state=name;

						break;
					}
					
					/* check if somesthing interesting happend */
					if (feof(sfp)!=0 || ferror(sfp)!=0 ) {
						(void)fclose(file_metadata.ffp);
						
						log_message(cmnd,verbose,FALSE,"Error while reading from stream (connection closed by peer/error)");
                        return FALSE;
                    }
				} while(1);
				log_verbose(cmnd,verbose, "Data received");
			}

		}

	}
	
	/* 
	 * we are here if fgets returned unsuccessfull
	 * this can happen in EOF or ERROR
	 */
	 
	/* everything went great! => return true */
	if (state==file && file_state==name) {
		log_verbose(cmnd,verbose, "Everything went fine. Connection succcessfully closed");
		return TRUE;
	}
	/* something went wrong */
	else {	
		log_message(cmnd,verbose,FALSE,"Error while reading from stream (connection closed by peer/error)");
		return FALSE;
	}

}

/*
 * --------------------------------------------------------------- helpers --
 */

/**
 * @brief logs messages if verbose flag is set
 *
 * Writes message to stderr using log_message if verboseflag is TRUE
 *
 * @param cmnd the command 
 * @param tmp_verbose verboseflag
 * @param useErrno the errno flag 
 * @param message 
 * @return void
 */	
void log_verbose(const char* cmnd, int* tmp_verbose, const char* message) {
	int verbose = *tmp_verbose;
	
	if (verbose == TRUE) {
		log_message(cmnd, tmp_verbose,FALSE, message);
	}
}

/**
 * @brief write a log message
 *
 * Writes the error message to stdout or stderr depending on verbose flag. 
 * With the useErrno flag it can even output the last error using strerror.
 *
 * @param cmnd the command 
 * @param tmp_verbose verboseflag
 * @param useErrno the errno flag 
 * @param message 
 * @return void
 */	
void log_message(const char* cmnd, int* tmp_verbose, int useErrno, const char* message) {
	FILE *stream=stderr;

	int verbose = *tmp_verbose;

	/* check if verbose */
	if (verbose == TRUE) {
		stream=stdout;
	}
	
	if (useErrno == FALSE) {
		fprintf(stream,"%s: %s\n",cmnd,message);
	}
	else {
		fprintf(stream,"%s: %s (%s)\n",cmnd,message,strerror(errno));
	}
}

/**
 * @brief prints a usage message
 *
 * Used as callback function for parsing the command line arguments. 
 * Is called on errors while parsing the command line args.
 *
 * @param stream the FILE* to write to
 * @param cmnd the command 
 * @param exitcode the supposed exit code 
 * @return void
*/	
void usage(FILE *stream, const char *cmnd, int exitcode) {
	fprintf(stream,"usage: %s options\n",cmnd);
	fprintf(stream,"options:\n");
	fprintf(stream,"-s, --server <server>   full qualified domain name or IP address of the server\n");
	fprintf(stream,"-p, --port <port>       well-known port of the server [0..65535]\n");
	fprintf(stream,"-u, --user <name>       name of the posting user\n");
	fprintf(stream,"-i, --image <URL>       URL pointing to an image of the posting user\n");
	fprintf(stream,"-m, --message <message> message to be added to the bulletin board\n");
	fprintf(stream,"-v, --verbose           verbose output\n");
	fprintf(stream,"-h, --help\n");
	
	exit(exitcode);
}

/*
 * =================================================================== eof ==
 */
