/**
 * @mainpage Example Program for TCP/IP Programming
 * @author florian.bogner@technikum-wien.at
 * @author michal.piatek@technikum-wien.at
 */

/**
 * @file simple_message_server_1384.c
 * @brief the server of the TCP/IP programming example
 */

/*
 * -------------------------------------------------------------- includes --
 */
 
#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>
#include <string.h>
#include <libgen.h>

/*
 * --------------------------------------------------------------- defines --
 */
 
/* default defines */
#define TRUE (1==1)
#define FALSE (!TRUE)

/* backlog value used for server socket */
#define BACKLOG 10

/* path to external server logic */
#define SERVER_LOGIC_CMND "/usr/local/bin/simple_message_server_logic"

/*
 * -------------------------------------------------------------- typedefs --
 */

/* typedef of callback from commandline parser */
typedef void (* sms_usagefunc_t) (FILE *, const char *, int);

/*
 * ------------------------------------------------------------ prototypes --
 */

void server_logic(int cd, const char* cmnd, int* verbose);
void sigchld_handler(int s);
void log_verbose(const char* cmnd, int* tmp_verbose, const char* message);
void log_message(const char* cmnd, int* tmp_verbose,int useErrno, const char* message);
void usage(FILE *stream, const char *cmnd, int exitcode);
void sms_parsecommandline(int argc, const char* const argv[], sms_usagefunc_t usagefunc ,const char **port, int* verbose);

/*
 * ------------------------------------------------------------- functions --
 */

/**
 * @brief the main function, starting point
 *
 * The main function parses the programm arguments, tries to sets up the server socket
 * and executes the server_logic function for each client after forking
 * 
 * @param  argc argument counter
 * @param  argv[] The arguments passed to the main function.
 * @return EXIT_SUCCESS (0)
 * @return EXIT_FAILURE (1)
 */
int main( int argc, const char* argv[] ) {
	/* local variables */
	int result;
	int sd,cd;
	socklen_t sin_size;
	struct addrinfo hints;
	struct addrinfo *server_addr_info, *p;
	struct sigaction sa;	
	struct sockaddr_storage client_addr;

	const char* port = NULL;
	int yes = 1;
	int verbose = FALSE;

	/* parse commandline - arguments have to be initialized with NULL */
	sms_parsecommandline(argc, argv, &usage, &port, &verbose); 
	
	/* configure hits structure for getaddrinfo (for prefering IPv4) */
	log_verbose(argv[0],&verbose, "Configure hints structure for getaddrinfo");
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	
	/* 
	 * try to create valid addr_info structure with valid sockets
	 * while parsing port information and hints 
	 */
	log_verbose(argv[0],&verbose, "Use getaddrinfo to get possible addresses");
	result = getaddrinfo(NULL, port, &hints, &server_addr_info);
	if (result!=0) {	
		log_message(argv[0],&verbose,FALSE,gai_strerror(result));
		
		return EXIT_FAILURE;
	}

	/* loop through all possible addresses and try to bind */ 
	log_verbose(argv[0],&verbose, "Loop through all addresses and try to bind");
	for(p = server_addr_info; p != NULL; p = p->ai_next) {
		/* create socket */
		if ((sd = socket(p->ai_family, p->ai_socktype, p->ai_protocol)) == -1) {
			log_message(argv[0],&verbose,TRUE,"Failed to create socket");
			continue;
		}
	
		/* enable socket reuse */
		if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
			log_message(argv[0],&verbose,TRUE,"Failed to set socket options");
			return EXIT_FAILURE;
		}

		/* and bind */
		if (bind(sd, p->ai_addr, p->ai_addrlen) == -1) { 
			close(sd);
			log_message(argv[0],&verbose,TRUE,"Failed to bind to socket");
			continue;
		} 

		break;
	}

	/* clean up getaddrinfo */
	log_verbose(argv[0],&verbose, "Clean up getaddrinfo");
	freeaddrinfo(server_addr_info);

	/* check if a valid socket has been found */
	if (p == NULL) {
		log_message(argv[0],&verbose,FALSE,"No socket found to bind to"); 
		return EXIT_FAILURE;
	}
	log_verbose(argv[0],&verbose, "A valid socket has been created");

	/* start listening */
	if (listen(sd, BACKLOG) == -1) { 
		log_message(argv[0],&verbose,TRUE,"Failed to listen");
		return EXIT_FAILURE;
	}
	log_verbose(argv[0],&verbose, "Listening started");
	
	/* configure signal handler, so that we can clean up our children */
	log_verbose(argv[0],&verbose, "Configuring signal handler");
	sa.sa_handler = sigchld_handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags=SA_RESTART;
	
	/* and register the handler */
	if (sigaction(SIGCHLD, &sa, NULL) == -1) {
		log_message(argv[0],&verbose,TRUE,"Failed to set sigchld handler");
		return EXIT_FAILURE;
	}

	/* main loop! -> wait for connections and fork! */
	log_verbose(argv[0],&verbose, "Starting main loop");
	while(1) {
		sin_size = sizeof(client_addr);
		/* wait for new client */
		log_verbose(argv[0],&verbose, "Waiting for new clients");
		cd = accept(sd, (struct sockaddr *)&client_addr, &sin_size);
		if (cd == -1) {
			log_message(argv[0],&verbose,TRUE,"Error while accepting new client");
			continue;
		}

		log_verbose(argv[0],&verbose, "Forking");
		result=fork();

		/* check if we are the child or the parent */
		switch(result) {
			case -1:
				log_message(argv[0],&verbose,TRUE,"Error while closing forking");
				(void)close(sd);
				(void)close(cd);
				
				return EXIT_FAILURE;
				
				break;
			case 0:
				/* child => execute server_logic */
				if (close(sd) != 0) {
					log_message(argv[0],&verbose,TRUE,"Error while closing client socket after fork");
					(void)close(cd);
					return EXIT_FAILURE;
				}
				
				log_verbose(argv[0],&verbose, "I'm the child: starting the logic");
				server_logic(cd, argv[0], &verbose);

				break;
			default: 
				/* parent => keep listening on server socket */
				if (close(cd) != 0) {
					log_message(argv[0],&verbose,TRUE,"Error while closing server socket after fork");
					(void)close(sd);
					return EXIT_FAILURE;
				}
				
				log_verbose(argv[0],&verbose, "I'm the parent: keeping listening");
				
				break;
		}
	}	
	
	/* close and exit (even if it will never happen) */
	(void)close(sd);
	
	return EXIT_SUCCESS;
}

/**
 * @brief executes the external server logic
 *
 * Redirects stdin and stdout to the given socket and then executes the server logic defined in SERVER_LOGIC_CMND
 * 
 * @param cd the client socket 
 * @param cmnd command name used for logging
 * @param verbose the verbose flag (determines where to write the error output)
 * @return void
 */
void server_logic(int cd, const char* cmnd, int* verbose) {
	int result;
	char* slc_copy;
	char* arg0;

	/* redirect stdout and stdin */
	log_verbose(cmnd,verbose, "Redirecting stdin and stdout; Starting server logic");
	if ( dup2(cd, STDIN_FILENO) == -1 || dup2(cd, STDOUT_FILENO) == -1 ) {
        log_message(cmnd,verbose,TRUE,"Error while redirecting stdin and stdout");
		(void)close(cd);
		exit(EXIT_FAILURE);;
	} 

	if (close(cd) != 0) {
		log_message(cmnd,verbose,TRUE,"Error while closing client socket after redirecting");
        exit(EXIT_FAILURE);
	}

	slc_copy = strdup(SERVER_LOGIC_CMND);
	if (slc_copy == NULL) {
		log_message(cmnd,verbose,FALSE,"Error while duplicating SERVER_LOGIC_CMND");
        exit(EXIT_FAILURE);
	}
	
	arg0 = basename(slc_copy);	

	result = execl(SERVER_LOGIC_CMND, arg0, NULL);
	
	/* shouldn't be reached */
	log_message(cmnd,verbose,TRUE,"Error while executing SERVER_LOGIC");
	(void)close(cd);
	exit(EXIT_FAILURE);
}

void sigchld_handler(int s) {
	s=s; /* hide unused parameter warning */
	while (waitpid(-1, NULL, WNOHANG) > 0);
}

/*
 * --------------------------------------------------------------- helpers --
 */

/**
 * @brief logs messages if verbose flag is set
 *
 * Writes message to stderr using log_message if verboseflag is TRUE
 *
 * @param cmnd the command 
 * @param tmp_verbose verboseflag
 * @param useErrno the errno flag 
 * @param message 
 * @return void
 */	
void log_verbose(const char* cmnd, int* tmp_verbose, const char* message) {
	int verbose = *tmp_verbose;
	
	if (verbose == TRUE) {
		log_message(cmnd, tmp_verbose, FALSE, message);
	}
}

/**
 * @brief write a log message
 *
 * Writes the error message to stdout or stderr depending on verbose flag. 
 * With the useErrno flag it can even display the last error using strerror.
 *
 * @param cmnd the command 
 * @param tmp_verbose verboseflag
 * @param useErrno the errno flag 
 * @param message 
 * @return void
 */	
void log_message(const char* cmnd, int* tmp_verbose, int useErrno, const char* message ) {
	FILE *stream=stderr;

	int verbose = *tmp_verbose;

	/* check if verbose */
	if (verbose == TRUE) {
		stream=stdout;
	}
	
	if (useErrno == FALSE) {
		fprintf(stream,"%s: %s\n",cmnd,message);
	}
	else {
		fprintf(stream,"%s: %s (%s)\n",cmnd,message,strerror(errno));
	}
}

/**
 * @brief prints a usage message
 *
 * Used as callback function for parsing the command line arguments. 
 * Is called on errors while parsing the command line args.
 *
 * @param stream the FILE* to write to
 * @param cmnd the command 
 * @param exitcode the supposed exit code 
 * @return void
 */	
void usage(FILE *stream, const char *cmnd, int exitcode) {
	fprintf(stream,"usage: %s option\n",cmnd);
	fprintf(stream,"options:\n");
	fprintf(stream,"	-p, --port <port>\n");
	fprintf(stream,"	-v, --verbose\n");
	fprintf(stream,"	-h, --help\n");

	exit(exitcode);
}

/**
 * @brief parses the command line args
 *
 * Tries to parse the arguments from argv to the given variables. On error (or usage) the usage function is called
 *
 * @param argc commandline count
 * @param argv commandline args
 * @param usagefunc error and usage function
 * @param port where to store the port number
 * @param verbose where to store verbose flag
 * @return void
 */	
void sms_parsecommandline(int argc, const char* const argv[], sms_usagefunc_t usagefunc ,const char **port, int* verbose) {
	int c;
	
	struct option long_options[] = {
		{"port",1,NULL,'p'},
		{"verbose",0,NULL,'v'},
		{"help",0,NULL,'h'},
		{0,0,0,0}
	};

	while( (c = getopt_long(argc, (char ** const) argv, "p:hv", long_options, NULL)) != -1) {

		switch (c) {
			case 'p':
				*port = optarg;
				break;
			case 'v':
				*verbose = TRUE;
				break;
			case 'h': 
				usagefunc(stdout, argv[0], EXIT_SUCCESS);
				break;
			case '?':
			default:
				usagefunc(stderr, argv[0], EXIT_FAILURE);
				break;
		}
	}
	
	if (optind != argc || *port==NULL) {
		usagefunc(stderr, argv[0], EXIT_FAILURE);
	}
}

/*
 * =================================================================== eof ==
 */
